
/*			DISPLAY for the CFS-C Classifier System

This file, DISPLAY.C, is part of the CFS-C classifier system (Copyright 1986,1988 Rick L. Riolo).

The subroutines in this file are called to DISPLAY variables, help, etc.
The subroutines:

	Display 	main entry point for DISPLAY command.
	DsplVar 	display a variable.
	AuDisplay	main entry for auto-display mechanism.
	DsplVars	| display selected variables.
	PrnVars 	|
	WrtVars 	|
	DsplSets	Display special settings, constants, etc.
	Help		The HELP command text.
**/

#include	"config.h"

#include	"core.ext"
#include	"utility.h"
#include	"utility.ext"
#include	"cfsio.ext"
#include	"cfmsgio.ext"
#include	"cfsutil.ext"
#include	"usercmd.ext"
#include	"display.def"
#include	"cfscstruct_io.h"   /* For PARALLIFE-like variable - I/O */

#if MPWC
#define __SEG__ CFSCWRITE
#endif


/*******************************

Display 	Called when user enters a DISPLAY command.
	
CmdPars 	A character array buffer that contains any parameters specified on the DISPLAY command.

The parameters have been converted to lowercase, and the string is
terminated by a '\0'. Each parameter in CmdPars is of the form:
	what[,format[,where]]
Each parameter is separated from the next by one or more blanks.
Display() picks off each parameter until none are left. For each one,
it calls approriate function to display it (or prints error message if unknown 'what').

******/

VOID 
DoDisplay (CmdPars)
    char CmdPars[];
{
    int formati = DFLTFMT;		/* int form of display forrmat */
    char par[PARNAMSZ + PARVALSZ + 1];	/* room for parameter and \0 */
    char name[PARNAMSZ + 1];		/* room for name part and \0 */
    char format[PARNAMSZ + 1];		/* room for the format part */
    char where[PARNAMSZ + 1];		/* room for where part */
    char *parptr, *lineptr;		/* pointer to next part of Par; Pointer to next par to fetch from CmdPars */
    extern char Version[], RunDate[], RunTime[];  /* Ok, they stay in cfsc.c and are only used there and here  (J.Engel) */
    extern int DsplEnv();

    for (lineptr = CmdPars; TRUE;) {
	lineptr = StpTok2 (lineptr, par, sizeof (par), " ");	/* Pick off next par, if any */
	if (par[0] == '\0')
	    break;			/* None found -- all done */
	else {				/* Get the various parts of this parameter from par variable. */
	    parptr = StpTok2 (par, name, sizeof (name), ",");
	    parptr = StpTok2 (parptr, format, sizeof (format), ",");
	    parptr = StpTok2 (parptr, where, sizeof (where), " ");
	    if (format[0] != '\0') {
		if (!IsInt (format))
		    WriteStd ("\nDISPLAY 'format' value should be an integer.\n");
		else
		    formati = atoi (format);
	    }
	    if (strcmp (name, "cl") == 0)
		DsplCfs (where, formati);

	    else if (strcmp (name, "ml") == 0)
		DsplMsgs (where, formati, 'c');

	    else if (strcmp (name, "oml") == 0)
		DsplMsgs (where, formati, 'o');

	    else if (strcmp (name, "env") == 0)
		DsplEnv (where, formati);

	    else if (strcmp (name, "vars") == 0)
		DsplVars (where, formati);

	    else if (strcmp (name, "settings") == 0)
		DsplSets (where, formati);

	    else if (strcmp (name, "cyclestp") == 0 || strcmp (name, "step") == 0) {
		sprintf (GOutBuff, "\nNext Major-Cycle Step:   %u.\n", CycleStp);
		WriteS (GOutBuff, DisOFILE);
	    } else if (strcmp (name, "banner") == 0) {
		sprintf (GOutBuff, "\nCFC-C Version %s.	 Run-Date  %s  (%s).\n", Version, RunDate, RunTime);
		WriteS (GOutBuff, DisOFILE);
	    } else if (strcmp (name, "time") == 0) {
		WriteS ("\nTime:  ", DisOFILE);
		GetCFSCTime (GOutBuff, 1);	/* hh:mm:ss form */
		*(GOutBuff + 8) = '\n';
		*(GOutBuff + 9) = '\0';
		WriteS (GOutBuff, DisOFILE);
	    } else
		DsplVar (name, where);	/* format ignored */
	}
    }

}					/* Display */


/*******************************

DsplVar 	Display value of specific variable in the cfsc_s structure (see cfsctypes.h).

	VarName	Name of variable to display.

	FDName	String name of File/device.
			(Default for "" is DisOutFN--the display output stream.)

******/

int 
DsplVar (VarName, FDName)
    char *VarName, *FDName;
{
    char *cp;
    long l,len;
    FILE *fileptr;

    while (*VarName == ' ')
	++VarName;			/* skip leading blanks, and cut off trailors. */
    for (cp = VarName; *cp != ' ' && *cp != '\0'; ++cp);
    if (*cp == ' ')
	*cp = '\0';

    while (*FDName == ' ')
	++FDName;

    if (strcmp (FDName, DisOutFN) == 0 || *FDName == '\0')
	fileptr = DisOFILE;		/* DisOutFN is already open */
    else if (strcmp (FDName, StdOutFN) == 0)
	fileptr = StdOFILE;		/* StdOutFN is already open */
    else if ((fileptr = GetFILE (FDName, "", WRITEMODE)) == NULL)
	return (ERROR);

    PrnVar(VarName,fileptr);

    if (fileptr != DisOFILE && fileptr != StdOFILE)	/* Dont close StdOutFN stream! */
	CloseFILE (FDName, "", fileptr);

    return (OK);

}					/* DsplVar */

/*******************************

PrnVar		Print a variable into an opened file / device:

****/
VOID
PrnVar(char *VarName, FILE *FilePtr)
{
	char localbuffer[80];
	long len,l;

	len=cfsc_sEntry(-1,localbuffer);
	strcpy(localbuffer,VarName);
	for (l=0;l<len;l++) {  /* look for variable in structure */
	    cfsc_sEntry(l,GOutBuff);
	    if (strcasecmp(localbuffer,GOutBuff)==0) 
		break;
	}
	if (l==len) {  /* not found */
	    sprintf (GOutBuff, "\nVariable '%s' not found.", VarName);
	    WriteStd (GOutBuff);
	    WriteS (GOutBuff, FilePtr);
	} else {  /* found index 'l' of variable */
	    strcat(GOutBuff,"=");
	    cfsc_sOut((char *)CFSC_VAR,l,GOutBuff+strlen(GOutBuff),GOBUFFSZ-strlen(GOutBuff));
	    WriteS (GOutBuff, FilePtr);
	    WriteS ("\n", FilePtr);
	}
}

/*******************************

AuDisplay	Auto-display various items user wants to see.

This is called from Classify() when AutoDspl > 0, i.e., when the user
wants one or items automatically displayed every specified number of steps.

Check each ADxxxInt variable: if CycleStp mod ADxxxInt = 0 then display the item 'xxx'.
Display item xxx in format specified by ADxxxFmt variable

NOTE: Detector messages are auto-displayed in RdDetect (since this function, AuDisplay,
is called at the bottom of the Major cycle before there are any detector messages.

******/

VOID 
AuDisplay ()
{
    if (CycleStp % ADCycInt == 0) {
	sprintf (GOutBuff, "%\n**** At end of Cycle Step %u ****\n", CycleStp);
	WriteS (GOutBuff, DisOFILE);
    }
    if (CycleStp % ADCMInt == 0)
	DsplMsgs ("", ADCMFmt, 'c');

    if (CycleStp % ADCfInt == 0)
	DsplCfs ("", ADCfFmt);

    if (CycleStp % ADVarInt == 0)
	DsplVars ("", ADVarFmt);

    if (CycleStp % ADEnvInt == 0)
	DsplEnv ("", ADEnvFmt);

}					/* AuDisplay */


/*******************************

DsplVars	Display system variables.

	FDName	string name of output file/device.
			(Default for "" is to write to DisOutFN--the display output stream.)

	Format	parameter determines the format in which variables are to be displayed.
			See WrtMsgs (below) for Format values and what they do.

DsplVars just opens the file, calls WrtVars or PrnVars to do the work, and closes the file.
**/

int 
DsplVars (FDName, Format)
    char *FDName;
    int Format;
{
    long l,len;
    FILE *fileptr;

    if (Format == DSPLOFF)
	return (OK);

    while (*FDName == ' ')
	++FDName;

    if (strcmp (FDName, DisOutFN) == 0 || *FDName == '\0')
	fileptr = DisOFILE;		/* DisOutFN is already open */
    else if (strcmp (FDName, StdOutFN) == 0)
	fileptr = StdOFILE;		/* StdOutFN is already open */
    else if ((fileptr = GetFILE (FDName, "", WRITEMODE)) == NULL)
	return (ERROR);

    /* There is only one format of variable output, so 'Format' is ignored.    (J.E.)  */
    len=cfsc_sEntry(-1,GOutBuff);
    for (l=0;l<len;l++) {
	cfsc_sEntry(l,GOutBuff);
	PrnVar(GOutBuff,fileptr);
    }

    if (fileptr != DisOFILE && fileptr != StdOFILE)	/* Dont close StdOutFN stream! */
	CloseFILE (FDName, "", fileptr);

    return (OK);

}					/* DsplVars */


/*******************************

PrnVars 	Write variables to open file/device.

	FilePtr	Opened file/device.

	Format	Select which variables to print.

******/
VOID 
PrnVars (FilePtr, Format)
    FILE *FilePtr;
    unsigned int Format;
{
    long len,l;

	len=cfsc_sEntry(-1,GOutBuff);
	for (l=0;l<len;l++) {
	    cfsc_sEntry(l,GOutBuff);
	    PrnVar(GOutBuff,FilePtr);
	}
}

/*******************************

WrtVars 	Write variables to already opened file/device.
	This is skipped due to the new variable-handling.

************/


/*******************************
DsplSets	 Display system control settings, constants, etc.

	The parameter specifies a format.

******/

int 
DsplSets (FDName, Format)
    char *FDName;
    int Format;
{
    FILE *destin;

    if (Format == DSPLOFF)
	return (OK);

    if (strcmp (FDName, DisOutFN) == 0 || *FDName == '\0')
	destin = DisOFILE;		/* DisOutFN is already open */
    else if (strcmp (FDName, StdOutFN) == 0)
	destin = StdOFILE;		/* StdOutFN is already open */
    else if ((destin = GetFILE (FDName, "", WRITEMODE)) == NULL)
	return (ERROR);

    WriteS ("\n\nSystem Settings:\n", destin);

    WriteS ("\nDefault I/O files/devices:", destin);
    sprintf (GOutBuff, "\n  Load messages from:	 %s", MsgInFN);
    WriteS (GOutBuff, destin);
    sprintf (GOutBuff, "\n  Load classifiers from:  %s", CfInFN);
    WriteS (GOutBuff, destin);
    sprintf (GOutBuff, "\n  Load environment from:  %s", EnvInFN);
    WriteS (GOutBuff, destin);

    sprintf (GOutBuff, "\n  Standard output on:	 %s", StdOutFN);
    WriteS (GOutBuff, destin);
    sprintf (GOutBuff, "\n  Display output on:	  %s", DisOutFN);
    WriteS (GOutBuff, destin);
    WriteS ("\n  Log file:			", destin);
    if (LogFName[0] == '\0')
	WriteS ("No log file.", destin);
    else {
	sprintf (GOutBuff, "%s", LogFName);
	WriteS (GOutBuff, destin);
    }
    sprintf (GOutBuff, "\n  Save-system on:		%s%s.%03d\n", SaveSyPa, SaveSyFN, SaveFNum);
    WriteS (GOutBuff, destin);

    if (destin != DisOFILE && destin != StdOFILE)	/* Don't close these streams! */
	CloseFILE (FDName, "", destin);

    return (OK);

}					/* DsplSets */


/*******************************

Help -- the HELP command.
	ParStr - indicate what display:
	 \0 	 - display command list.
	 set	 - SET command parameters.
	 display - DISPLAY command parameters.

******/

VOID 
Help (ParStr)
    char ParStr[];
{

    WriteStd ("\nSorry, no help.\n");     /* ?!? */

#if ( MPWC || SUN3 || MACLSC || APOLLOC || VAX )
    if (ParStr[0] == '\0') {
	WriteStd ("\nCommands and parameters");
        WriteStd ("\nTo stop:                                   STOP");
        WriteStd ("\nTo load messages:                          { LOADC | LC } [ file-name ]");
        WriteStd ("\nTo load classifiers                        { LOADM | LM } [ file-name ]");
        WriteStd ("\nTo load the 'environment':                 { LOADE | LE } [ file-name ]");
        WriteStd ("\nTo display things:                         { DISPLAY | DI } item [ ,item ]*");
        WriteStd ("\nTo set parameters                          SET item=value [ ,item=value ]*");
        WriteStd ("\nTo run the system n steps / 1 step:        { CLASSIFY | C } n  / { STEP | ST } ");
        WriteStd ("\nTo raise the XView window:                 WIN");
        WriteStd ("\nTo execute an enviroment command:          ECMD  command");
        WriteStd ("\nTo read commands from a file:              { READCMDS | RCS } file");
        WriteStd ("\nTo save classifiers for reload:            DISPLAY cl,filename,99");
        WriteStd ("\nTo save messages for reload:               DISPLAY ml,filename,99");
        WriteStd ("\nTo save the entire system state:           SAVESYS  [filename]");
        WriteStd ("\nTo append classifiers to list:             APPC  [ file-name ]");
        WriteStd ("\nTo append messages to list:                APPM  [ file-name ]");
        WriteStd ("\nTo modify a classifier:                    MC      id  (not implemented)");
        WriteStd ("\nTo modify a message:                       MM      id  (not implemented)");
        WriteStd ("\nHELP                                       HELP [ { SET | DISPLAY } ]");
	WriteStd ("\nUppercase means enter as shown. 'file-name' can be a path (if supported).");
	WriteStd ("\n[ ] indicates optional; { X | Y } indicate alternative forms.");
    } else if (strcmp (ParStr, "set") == 0) {
	WriteStd ("\nCommon SET-able items:");

	WriteStd ("\nTo control list sizes, taxes and other accounting:");
	WriteStd ("\n  nms=n	  -- maximum number of new (internally generated) messages.");
	WriteStd ("\n  ncfs=n	 -- maximum number of classifiers");

	WriteStd ("\n  strdf=d.d  -- default classifier strength.");
	WriteStd ("\n  strmax=d.d -- maximum classifier strength.");
	WriteStd ("\n  intdf=d.d  -- default intensity for messages (from detectors).");
	WriteStd ("\n  supdf=d.d  -- default support for classifiers (for NOT-conditions).");

	WriteStd ("\n  bidk=d.d   -- bid 'risk' factor.");
	WriteStd ("\n  bidmin=d.d -- minimum bid possible.");
	WriteStd ("\n  htax=d.d   -- head tax.");
	WriteStd ("\n  btax=d.d   -- bid  tax.");
	WriteStd ("\n  ptax=d.d   -- producer tax.");

	WriteStd ("\n  paydet=d.d -- fraction of bid to pay for detector messages");

	WriteStd ("\nTo control of detector/effector interface sampling rates:");
	WriteStd ("\n  drate=n	-- read messages from detectors every n steps.");
	WriteStd ("\n  erate=n	-- activate matched effectors every n steps.");

	WriteStd ("\nTo control Output:");
	WriteStd ("\n  echo={0|1}		 -- echo commands (for use when input is from a file).");
	WriteStd ("\n  log={on|off|file}  -- log I/O to a file (** not implemented yet **).");
	WriteStd ("\n  demo=n			 -- set demonstration output 'level' to n, where: ");
	WriteStd ("\n						  0 = Off	   1 = Summaries ");
	WriteStd ("\n						  2 = Details from 'major cycle' step.  (99 = DEBUG).");
	WriteStd ("\n						Note that this resets the dml, dcl, and denv flags.");
	WriteStd ("\n  di=n			   -- (auto)display interval, i.e., items with auto-display ");
	WriteStd ("\n						set > 0 displayed every n steps.");
	WriteStd ("\n  The auto-displayable items are:");
	WriteStd ("\n	dml={0|format}	 -- autodisplay message list.");
	WriteStd ("\n	dcl={0|format}	 -- autodisplay classifier list. ");
	WriteStd ("\n	dv={0|format}	  -- autodisplay variables.");
	WriteStd ("\n	denv={0|format}	-- autodisplay environment state.");

	WriteStd ("\nOther utility controls (for re-starting runs, etc.):");
	WriteStd ("\n  cstep=n	-- cycle step (major cycle clock).");
	WriteStd ("\n  randseed=n -- seed for pseudo-random number generator.");

	WriteStd ("\n  lmf=file   -- default file for LOADM (load messages) command.");
	WriteStd ("\n  lcf=file   -- default file for LOADC (load classifiers) commmand.");
    } else if (strcmp (ParStr, "display") == 0) {
	WriteStd ("\nDISPLAY-able items. Each item should be of the form:");
	WriteStd ("\n	  item[,format[,fname]]");
	WriteStd ("\nwhere 'format' is an integer from 0..99 which specifies the");
	WriteStd ("\nlevel of detail and/or format of the display, and where");
	WriteStd ("\n'fname' is the name of a destination file. The items:\n");

	WriteStd ("\n  env		  state of 'environment'");
	WriteStd ("\n  ml		   current message list");
	WriteStd ("\n  cl		   current classifier list");
	WriteStd ("\n  vars		 values of system variables");
	WriteStd ("\n  settings	 values of system settings, constants, etc. ");
	WriteStd ("\n  time		 current time.");
	WriteStd ("\n  banner	   CFS-C version and date and time of run.");
	WriteStd ("\n\nBy default, items are displayed on the screen (stdout)");
	WriteStd ("\nin a format deemed 'reasonable' for the item.");
    } else
	WriteStd ("\n Illegal parameter on HELP command. Enter HELP for help.");

    WriteStd ("\n");
#endif

}					/* Help */
